#!/bin/bash
# first argument is the file with pointers to image
# s
bash run_with_bash_copy_files_according2txt.sh ../pacs_split/art_painting_10.txt ~/Documents/dataset/PACS/raw .
bash run_with_bash_copy_files_according2txt.sh ../pacs_split/cartoon_10.txt ~/Documents/dataset/PACS/raw .
bash run_with_bash_copy_files_according2txt.sh ../pacs_split/photo_10.txt ~/Documents/dataset/PACS/raw .
bash run_with_bash_copy_files_according2txt.sh ../pacs_split/sketch_10.txt ~/Documents/dataset/PACS/raw .
